﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Xml.Serialization;

namespace BlockPuzzleSolution
{
    public class Puzzle
    {
        public List<Block> Blocks;
        public Size PuzzleSize;
        public int[] Matrix;
        public bool AllowRotation;

        public int readMatrix(int x, int y)
        {
            if ((x < 0) || (y<0) || (x>=PuzzleSize.Width) || (y>=PuzzleSize.Height) ) return 1;
            return Matrix[y * PuzzleSize.Width + x];
        }

        public int readOrder(int x, int y)
        {
            return readMatrix(x - 1, y) + readMatrix(x + 1, y) + readMatrix(x, y - 1) + readMatrix(x, y + 1);
        }

        public void setMatrix(int x, int y, int v)
        {
            Matrix[y * PuzzleSize.Width + x] = v;
        }

        public Puzzle()
        {
            Blocks = new List<Block>();
            AllowRotation = false;
        }

        public void initPuzzle(int w, int h)
        {
            PuzzleSize = new Size(w, h);
            AllowRotation = false;
            Matrix = new int[PuzzleSize.Width*PuzzleSize.Height];
            for (int i = 0; i < PuzzleSize.Width; i++)
                for (int j = 0; j < PuzzleSize.Height; j++)
                    setMatrix(i, j, 0);
        }



        private bool TestBlock(Block b, Point p)
        {
            for(int i=0; i < b.BlockSize.Width; i++)
                for (int j = 0; j < b.BlockSize.Height; j++)
                {
                    if (b.readMatrix(i, j) == 1 && readMatrix(p.X + i, p.Y + j) == 1) return false;
                }

            return true;
        }

        private void PutBlock(Block b, Point p)
        {
            for (int i = 0; i < b.BlockSize.Width; i++)
                for (int j = 0; j < b.BlockSize.Height; j++)
                {
                    if (b.readMatrix(i, j) == 1) setMatrix(p.X + i, p.Y + j, 1);
                }
            b.Position = p;
        }

        private void RemoveBlock(Block b)
        {
            if (b.Position.X < 0) return;
            for (int i = 0; i < b.BlockSize.Width; i++)
                for (int j = 0; j < b.BlockSize.Height; j++)
                {
                    if (b.readMatrix(i, j) == 1) setMatrix(b.Position.X + i, b.Position.Y + j, 0);
                }
            b.Position = new Point(-1, -1);
        }

        public void ClearSolution()
        {
            foreach (Block b in Blocks) b.Position = new Point(-1, -1);
            for (int i = 0; i < this.Matrix.Length; i++) this.Matrix[i] = 0;
        }

        // This is the improved algorithm. It tests difficult cells first so it would
        // terminate early for wrong solutions.
        // The performance is improved significantly over brute-force algorithm
        public bool FindSolution2()
        {
            //find empty cell with biggest order (most taken cellls around it)
            int biggestOrder = -1;
            Point cellPoint = new Point(-1, -1);
            for (int i=0; i<PuzzleSize.Width; i++)
                for (int j = 0; j < PuzzleSize.Height; j++)
                {
                    if (readMatrix(i, j)==0 && (biggestOrder < readOrder(i, j)))
                    {
                        cellPoint.X = i;
                        cellPoint.Y = j;
                    }
                }
            if (cellPoint.X == -1) return true;  // not empty cell, a solution is found

            //for each block 
            // for each rotation
            //  for each possible position
            //    test if the block can fit the empty cell
            int px, py;
            foreach (Block b in Blocks)
            {
                if (b.Position.X >= 0) continue; // block is used, try next block
                for (int r = 0; r < 4; r++)
                {
                    for (int i=0; i<b.BlockSize.Width; i++)
                        for (int j = 0; j < b.BlockSize.Height; j++)
                        {
                            if (b.readMatrix(i, j) == 0) continue; // must fill the empty cell
                            px = cellPoint.X - i; py = cellPoint.Y - j; // start point
                            if (px < 0 || py < 0) continue; // out of puzzle border
                            if ((px + b.BlockSize.Width) > PuzzleSize.Width) continue;
                            if ((py + b.BlockSize.Height) > PuzzleSize.Height) continue;
                            if (TestBlock(b, new Point(px, py)))
                            {
                                PutBlock(b, new Point(px, py));
                                if (FindSolution2()) return true;
                                RemoveBlock(b);
                            }
                        }// position loop
                    if (AllowRotation) b.Rotate();
                    else break; // no rotation
                }// rotation loop
            } // block loop
            return false;
        }

        // This is obsolete brute-force algorithm. It takes too long (days or months) to solve medium size puzzle with rotation
        public bool FindSolution()
        {
            //find an unused block
            Block blockToTest = null;
            foreach (Block b in Blocks)
            {
                if (b.Position.X == -1) { blockToTest = b; break; }
            }
            // all blocks are used, a solution is found
            if (blockToTest == null) return true;
            for (int r = 0; r < 4; r++)
            {
                for (int i = 0; i <= PuzzleSize.Width - blockToTest.BlockSize.Width; i++)
                    for (int j = 0; j <= PuzzleSize.Height - blockToTest.BlockSize.Height; j++)
                    {
                        if (TestBlock(blockToTest, new Point(i, j)))
                        {
                            // put the blcok at this position
                            PutBlock(blockToTest, new Point(i, j));
                                // and find the solution for remaining blocks
                                if (FindSolution()) { return true; }
                                // if no solution is found, try next position
                                RemoveBlock(blockToTest);
                        }
                    }
                if (AllowRotation) blockToTest.Rotate();
                else break;
            }
            // no solution is found
            return false;
        }

        public Bitmap GetImage(int count)
        {
            int myUnitLength = Block.unitLength * 3;
            Bitmap bBitMap = new Bitmap(PuzzleSize.Width * myUnitLength, PuzzleSize.Height * myUnitLength);
            Graphics objGraphics = Graphics.FromImage(bBitMap);
            objGraphics.FillRectangle(new SolidBrush(Color.Gray), 0, 0, PuzzleSize.Width * myUnitLength, PuzzleSize.Height * myUnitLength);
            for (int ic=0; ic<Blocks.Count && ic<count; ic++) 
            {
                Block b = Blocks[ic];
                Bitmap bTexture = new Bitmap(myUnitLength, myUnitLength);
                Graphics graTexture = Graphics.FromImage(bTexture);
                LinearGradientBrush brTexture = new LinearGradientBrush(new Rectangle(0, 0, myUnitLength, myUnitLength), Color.White, b.MyColor, LinearGradientMode.ForwardDiagonal);
                graTexture.FillRectangle(brTexture, 0, 0, myUnitLength, myUnitLength);
                graTexture.Flush();
                TextureBrush br = new TextureBrush(bTexture);
                if (b.Position.X >= 0 && b.Position.Y >= 0)
                {
                    Pen wPen = new Pen(Color.White, 2);
                    for (int i = 0; i < b.BlockSize.Width; i++)
                        for (int j = 0; j < b.BlockSize.Height; j++)
                        {
                            if (b.readMatrix(i, j) == 1)
                            {
                                objGraphics.FillRectangle(br, (i + b.Position.X) * myUnitLength, (PuzzleSize.Height - j - b.Position.Y-1) * myUnitLength, myUnitLength, myUnitLength);
                                // draw border
                                //left
                                if (i == 0 || b.readMatrix(i - 1, j) == 0)
                                {
                                    objGraphics.DrawLine(wPen, (i+b.Position.X) * myUnitLength, 
                                        (PuzzleSize.Height - j - b.Position.Y -1) * myUnitLength, 
                                        (i+b.Position.X) * myUnitLength, 
                                        (PuzzleSize.Height  - j - b.Position.Y) * myUnitLength );
                                }
                                //right
                                if (i == (b.BlockSize.Width - 1) || b.readMatrix(i + 1, j) == 0)
                                {
                                    objGraphics.DrawLine(wPen, (i + b.Position.X + 1) * myUnitLength , 
                                        (PuzzleSize.Height - j - b.Position.Y-1) * myUnitLength, 
                                        (i + b.Position.X + 1) * myUnitLength , 
                                        (PuzzleSize.Height - j - b.Position.Y) * myUnitLength );
                                }
                                //top
                                if (j == (b.BlockSize.Height - 1) || b.readMatrix(i, j + 1) == 0)
                                {
                                    objGraphics.DrawLine(wPen, (i + b.Position.X) * myUnitLength, 
                                        (PuzzleSize.Height - j - b.Position.Y - 1) * myUnitLength, 
                                        (i + b.Position.X + 1) * myUnitLength, 
                                        (PuzzleSize.Height - j - b.Position.Y - 1) * myUnitLength);
                                }
                                //bottom
                                if (j == 0 || b.readMatrix(i, j - 1) == 0)
                                {
                                    objGraphics.DrawLine(wPen, (i + b.Position.X) * myUnitLength, 
                                        (PuzzleSize.Height - j - b.Position.Y) * myUnitLength , 
                                        (i + 1 + b.Position.X) * myUnitLength , 
                                        (PuzzleSize.Height - j - b.Position.Y) * myUnitLength);
                                }

                            }
                        }
                }
            }
            objGraphics.Flush();
            return bBitMap;
        }
    }
}
